# fpgadbg_conv.py
# Copyright (C) 10.07.2006 Wojciech M. Zabolotny (wzab@ise.pw.edu.pl)
#
# The code for conversion of data received from the fpgadbg_core into LXT file
# which is suitable for viewing with gtkwave viewer
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import lxt

# Simple utility function to convert the integer into the "01" character string
def word2str(word,nbits):
    word=int(word)
    res=""
    for i in range(0,nbits):
        if word & 1:
            res="1"+res
        else:
            res="0"+res
        word=int(word/2)
    return res   

class fpgadbg_conv:
    def __init__(self,assignments, nbits=16, timestep=100, timescale=-10, filename="out.lxt"):
        # Set some parameters
        self.nbits=nbits
        self.timestep=timestep
        # Open the output file and setup the timing parameters
        self.fout=lxt.lt_init(filename)
        lxt.lt_set_timescale(self.fout,timescale)
        # Setup the conversion table
        conv_table=[]
        for sigdef in assignments:
            if len(sigdef) != 3 and len(sigdef)!=5:
                raise "Wrong length of signal assignment definition"
            msb=sigdef[0]
            lsb=sigdef[1]
            name=sigdef[2]
            if len(sigdef)==5:
                omsb=sigdef[3]
                olsb=sigdef[4]
            else:
                olsb=0
                omsb=msb-lsb
            # Create the signal definition in the LXT file
            s=lxt.lt_symbol_add(self.fout,name,0,omsb,olsb,lxt.LT_SYM_F_BITS)
            # Add the description to our internal conversion table
            conv_table.append((msb,lsb,s,name))
        # Convert the conversion table from list into tuple (to optimize accesses)
        self.conv_table=tuple(conv_table)
    
    def convert_data(self,i):
        # We convert all the data words into a single "01" character string
        bits=""
        #print i
        for j in range(i,i+self.words_per_sample):
            bits=word2str(self.data[j],self.nbits)+bits
        # Now we can write the data
        #print bits
        for j in range(0,len(self.conv_table)):
            ct=self.conv_table[j]
            dl=len(bits)-1
            smsb=dl-ct[0] # Position of the first bit
            slsb=dl-ct[1]+1 # Position of the last bit
            lxt.lt_emit_value_bit_string(self.fout,ct[2],0,bits[smsb:slsb])
            #print ct[3]+":"+bits[smsb:slsb]
        lxt.lt_set_time(self.fout,self.t)

    def conv(self,data):
        # Store the reference to the data into the object
        self.data=data

        # Constant length of the data word
        word_len=self.nbits

        # Read the parameters from the data
        log2samples=data[0]
        # The highest bit stores the information if the buffer 
        # was completely filled (so the start position must be calculated
        # from the "stop" position)
        filled_mask=1<<(self.nbits-1)
        filled = ((log2samples & filled_mask) != 0)
        log2samples &= ~filled_mask
        num_of_samples = 1<<log2samples;
        data_width=data[1]
        # Calculate the number of data words per sample:
        self.words_per_sample=int((data_width+word_len-1)/word_len)
        # Position of first data word in the data array
        if self.nbits >= 16:
            first_data = 4
            trig_pos=data[2]*self.words_per_sample+first_data
            stop_pos=data[3]*self.words_per_sample+first_data
        else:
            first_data = 6
            trig_pos=(data[2]+256*data[3])*self.words_per_sample+first_data
            stop_pos=(data[4]+256*data[5])*self.words_per_sample+first_data

        # Calculate the trigger position
        

        # Calculate the position of the last recorded sample

        # Position AFTER the last data word in the data array
        after_last_data=len(data)

        # The code below could be used to set time scale to 0 at the
        # trigger position. Unfortunately the LXT format does not allow to
        # use negative time values...
        # Calculate the trigger position relatively to the start of the data
        # (it is necessary to set the time scale to 0 at the trigger position)
        #
        #if filled:
        #    rel_trig_pos = post_trigger_samples-num_of_samples
        #else:
        #    rel_trig_pos = -data[2] 
        #self.t = rel_trig_pos*self.timestep
        #lxt.lt_set_time(self.fout,self.t)
        #
        # Now we can only set time to 0 at the begining of the LXT file
        self.t=0
        lxt.lt_set_time(self.fout,self.t)
        
        # If the buffer has not been filled, the data start at the first_data,
        # and ends at the stop_pos.
        # However if the buffer has been filled completely, then the data
        # start after the stop position, and then continue from the begining
        # of the buffer to the stop position
        # Now we can copy the data, starting from the stop position
        #
        # past_pos will contain the position with the oldest data (if the buffer
        # has been filled)
        past_pos=stop_pos+self.words_per_sample
        if filled:
           # If the buffer was filled, then the valid data start after the stop position
           for i in range(past_pos,after_last_data,self.words_per_sample):
             self.t+=self.timestep
             self.convert_data(i)
        # The buffer between first_data and past_pos always contains the valid data     
        for i in range(first_data,past_pos,self.words_per_sample):
           self.t+=self.timestep
           self.convert_data(i)
        #Finally we close the file
        lxt.lt_close(self.fout)
    
