-------------------------------------------------------------------------------
-- Title      : fpgadbg_uart - interface between the fpgadbg and UART
-- Project    : 
-------------------------------------------------------------------------------
-- File       : fpgadbg_uart.vhd
-- Author     : Wojciech M. Zabolotny
-- Company    : 
-- Created    : 2006-07-08
-- Last update: 2006-07-28
-- Platform   : 
-- Standard   : VHDL'87
-------------------------------------------------------------------------------
-- Description: 
-------------------------------------------------------------------------------
-- Copyright (c) 2006 Wojciech M. Zabolotny (wzab@ise.pw.edu.pl) 
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2006-07-08  1.0      wzab      Created
-------------------------------------------------------------------------------
--  
--
--  This program is free software; you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation; either version 2 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License
--  along with this program; if not, write to the Free Software
--  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--  IMPORTANT EXTENSION OF YOUR RIGHTS:
--  You may also link this code with any other VHDL, Verilog, EDIF or other similar
--  code and distribute the resulting FPGA core in a binary form (as a configuration
--  bitstream or programmed FLASH memory or programmed FPGA chips) without the need
--  to disclose sources of your design (unless the license for that other code
--  disallows that).

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
library work;

entity fpgadbg_uart is
  generic (
    brg_div     : integer := 50000000/16/115200;
    width       : integer := 35;
    log2samples : integer := 10);
  port (
    wr_clk  : in  std_logic;
    trigger : in  std_logic;
    data_in : in  std_logic_vector((width-1) downto 0);
    rs_rx   : in  std_logic;
    rs_tx   : out std_logic;
    --dbg     : out std_logic_vector(7 downto 0);
    xconf   : out std_logic_vector(2 downto 0);
    sys_clk : in  std_logic;            -- system clock
    nrst    : in  std_logic             -- reset (active low)
    );
end fpgadbg_uart;

architecture syn of fpgadbg_uart is

  component fpgadbg
    generic (
      outwidth    : integer;
      width       : integer;
      log2samples : integer);
    port (
      wr_clk          : in  std_logic;
      trigger         : in  std_logic;
      wr_init         : in  std_logic;
      data_in         : in  std_logic_vector((width-1) downto 0);
      post_tr_samples : in  integer range 0 to 65535;
      triggered       : out std_logic;
      completed       : out std_logic;
      rd_init         : in  std_logic;
      rd_finished     : out std_logic;
      rd_clk          : in  std_logic;
      out_data        : out std_logic_vector((outwidth-1) downto 0));
  end component;

  component uart
    generic (
      brg_div : integer);
    port (
      clk         : in  std_logic;
      din         : in  std_logic_vector(7 downto 0);
      dout        : out std_logic_vector(7 downto 0);
      rx_rd       : in  std_logic;
      tx_wr       : in  std_logic;
      nrst        : in  std_logic;
      tx_empty    : out std_logic;
      tx_finished : out std_logic;
      rx_full     : out std_logic;
      fr_err      : out std_logic;
      ovr_err     : out std_logic;
      tx          : out std_logic;
      rx          : in  std_logic);
  end component;

  signal wr_init, rd_init, rd_clk, rd_finished, triggered, completed : std_logic;
  signal rs_tx_empty, rs_wr, rs_rd, rs_dav                           : std_logic;
  signal post_tr_samples                                             : integer;
  signal post_tr_vec                                                 : std_logic_vector(15 downto 0);
  signal rs_din, rs_dout, out_data                                   : std_logic_vector(7 downto 0);

  type   T_CTRL_STATE is (ST_IDLE, ST_WAIT_ACQ, ST_SEND_DATA, ST_SEND_DATA_C0);
  signal ctrl_state : T_CTRL_STATE;
  
begin

  fpgadbg_1 : fpgadbg
    generic map (
      outwidth    => 8,
      width       => width,
      log2samples => log2samples)
    port map (
      wr_clk          => wr_clk,
      trigger         => trigger,
      wr_init         => wr_init,
      data_in         => data_in,
      post_tr_samples => post_tr_samples,
      triggered       => triggered,
      completed       => completed,
      rd_init         => rd_init,
      rd_finished     => rd_finished,
      rd_clk          => rd_clk,
      out_data        => out_data);

  uart_1 : uart
    generic map (
      brg_div => brg_div)
    port map (
      clk         => sys_clk,
      din         => rs_dout,
      dout        => rs_din,
      rx_rd       => rs_rd,
      tx_wr       => rs_wr,
      nrst        => nrst,
      tx_empty    => rs_tx_empty,
      tx_finished => open,
      rx_full     => rs_dav,
      fr_err      => open,
      ovr_err     => open,
      tx          => rs_tx,
      rx          => rs_rx);

  post_tr_samples <= conv_integer(post_tr_vec);
  -- The line below was used to debug the fpgadbg_uart itself...
  --dbg             <= (0 => rd_init, 1 => wr_init, 2 => triggered, 3 => completed, 4 => rd_finished, others => '0');
  -----------------------------------------------------------------------------
  -- Main process
  -----------------------------------------------------------------------------
  pmain : process (nrst, sys_clk)
  begin  -- process pmain
    if nrst = '0' then                  -- asynchronous reset (active low)
      rs_wr      <= '0';
      rd_clk     <= '0';
      rd_init    <= '0';
      wr_init    <= '0';
      rs_rd      <= '0';
      ctrl_state <= ST_IDLE;
    elsif sys_clk'event and sys_clk = '1' then  -- rising clock edge
      rs_wr  <= '0';
      rd_clk <= '0';
      rs_rd  <= '0';
      case ctrl_state is
        when ST_IDLE =>
          if rs_dav = '1' then
            -- The UART has received a character
            if rs_din(7 downto 6) = "00" then
              -- reset the fpgadbg core, and store the lower bits
              -- of post_tr_samples
              post_tr_vec(5 downto 0) <= rs_din(5 downto 0);
              wr_init                 <= '0';
              rd_init                 <= '0';
              rs_rd                   <= '1';
            elsif rs_din(7 downto 6) = "01" then
              -- write the higher bits of post_tr_samples
              post_tr_vec(11 downto 6) <= rs_din(5 downto 0);
              rs_rd                    <= '1';
            else                        -- rs_din(7)='1'
              -- store the most significant bits of post_tr_samples
              post_tr_vec(15 downto 12) <= rs_din(3 downto 0);
              -- next three bits may be used to switch the active configuration
              xconf                     <= rs_din(6 downto 4);
              -- start the data acquisition
              rd_init                   <= '0';
              wr_init                   <= '1';
              rs_rd                     <= '1';
              ctrl_state                <= ST_WAIT_ACQ;
            end if;
          end if;
        when ST_WAIT_ACQ =>
          -- we are waiting until the acquisition finishes
          if completed = '1' then
            -- all data are recorded, we can transmit them
            rd_init    <= '1';
            ctrl_state <= ST_SEND_DATA;
          end if;
        when ST_SEND_DATA =>
          if rd_finished = '1' then
            -- all data have been transfered, go to the IDLE STATE
            ctrl_state <= ST_IDLE;
          elsif rs_tx_empty = '1' then

            -- if transmitter is ready, send the next byte
            rs_wr      <= '1';
            rd_clk     <= '1';
            rs_dout    <= out_data;
            ctrl_state <= ST_SEND_DATA_C0;
          end if;
        when ST_SEND_DATA_C0 =>
          rd_clk     <= '0';
          rs_dout    <= out_data;
          ctrl_state <= ST_SEND_DATA;
        when others => null;
      end case;
    end if;
  end process pmain;

end syn;
